/*--------------------------------------------------------------------
 * $Id: wcloud3d.c,v 1.10 2003/07/11 17:09:42 pa2h Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <limits.h>
#include <string.h>
#include "ascii.h"
#include "wcloud3d.h"

#if HAVE_LIBNETCDF
#include <netcdf.h>
#endif

#ifndef PI
#define PI 3.14159265358979323846264338327
#endif

/* Maximum number of characters per line */
#define MAX_LENGTH_OF_LINE     65536


/*******************************************************/
/* Read data from a 3D cloud description file.         */
/*******************************************************/
 
int read_3D_cloud (char *filename, 
		   int *Nx, int *Ny, int *Nz,
		   double *delX, double *delY,
		   double  **zcld3D, float ****kcld3d, int **threed,
		   float ****gcld3d, int *gspec,
		   float ****wcld3d, int *wspec,
		   int quiet)
{
  int i=0, number=0, status=0;
  int ix=0, iy=0, iz=0;

  int rows=0, min_columns=0, max_columns=0, max_length=0;

  char line[MAX_LENGTH_OF_LINE+1]="";
  char *string=NULL;

  FILE *f=NULL;
  char **array=NULL;
  char *dummy=NULL;


  /* check input file */
  status =  ASCII_checkfile (filename, 
			     &rows,
			     &min_columns,
			     &max_columns,
			     &max_length);
  
  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n",
	     status, filename);
    return status;
  }
    
  if (rows<2) {
    fprintf (stderr, "Error: found less than two rows in %s\n", filename);
    return -1;
  }

  if (min_columns<3) {
    fprintf (stderr, "Error: found less than three columns in %s\n", filename);
    return -1;
  }
  
  /* reset string to the beginning of the memory block */
  string = line;
  
  /* open file */
  if ( (f = fopen(filename, "r")) == NULL)  
    return ASCIIFILE_NOT_FOUND;
  

  /* first line */
  number=0;
  while (number==0) {
    fgets (string, MAX_LENGTH_OF_LINE, f);
    status = ASCII_parsestring (string, &array, &number);

    if (status!=0) {
      fprintf (stderr, "Error %d reading 1st line of %s\n", status, filename);
      return status;
    }
  }

  if (number<3) {
    fprintf (stderr, "Error: expected at least 3 values in 1st line, found only %d\n", number);
    return -1;
  }
  
  /* first 3 values are number of cells */
  *Nx = strtol (array[0], &dummy, 0);
  *Ny = strtol (array[1], &dummy, 0);
  *Nz = strtol (array[2], &dummy, 0);

  if (number>3)
    *gspec = strtol (array[3], &dummy, 0);
    
  if (number>4)
    *wspec = strtol (array[4], &dummy, 0);
  
  free (array);

  if (!quiet) {
    fprintf (stderr, " ... reading %d x %d x %d data points from %s\n",
	     *Nx, *Ny, *Nz, filename);
    fprintf (stderr, " ... gspec = %d, wspec = %d\n",
	     *gspec, *wspec);
  }

  /* 2nd line */
  number=0;
  while (number==0) {
    fgets (string, MAX_LENGTH_OF_LINE, f);
    status = ASCII_parsestring (string, &array, &number);

    if (status!=0) {
      fprintf (stderr, "Error %d reading 1st line of %s\n", status, filename);
      return status;
    }
  }

  if (number!=*Nz+1+2) {
    fprintf (stderr, "Error: found %d z levels, expected %d\n",
	     number-2, *Nz+1);
    return -1;
  }

  /* horizontal distances, convert from km to m */
  *delX = strtod (array[0], &dummy) * 1000.0;
  *delY = strtod (array[1], &dummy) * 1000.0;

  if (!quiet) {
    fprintf (stderr, " ... x-distance: %g\n", *delX);
    fprintf (stderr, " ... y-distance: %g\n", *delY);
  }

  /* allocate memory for altitude levels */
  *zcld3D = calloc ((size_t) *Nz+1, sizeof(double));
  if (*zcld3D==NULL) {
    fprintf (stderr, "Error allocating memory for zcld3D\n");
    return -1;
  }
	     
  *threed = calloc ((size_t) *Nz, sizeof(int));
  if (*threed==NULL) {
    fprintf (stderr, "Error allocating memory for threed\n");
    return -1;
  }
	     
  /* set altitude levels and convert from km to m */
  for (i=2; i<=*Nz+2; i++)
    (*zcld3D)[i-2] = strtod(array[i], &dummy)*1000.0;

  free(array);

  


  /* allocate memory for optical depth, asymmetry factor, and single scattering albedo */
  *kcld3d = calloc((size_t) *Nz, sizeof(float **));

  if (*gspec)
    *gcld3d = calloc((size_t) *Nz, sizeof(float **));

  if (*wspec)
    *wcld3d = calloc((size_t) *Nz, sizeof(float **));


  for (iz=0; iz<*Nz; iz++) {
    (*kcld3d)[iz] = calloc((size_t) *Nx, sizeof(float *));

    if (*gspec)
      (*gcld3d)[iz] = calloc((size_t) *Nx, sizeof(float *));

    if (*wspec)
      (*wcld3d)[iz] = calloc((size_t) *Nx, sizeof(float *));


    for (ix=0; ix<*Nx; ix++) {
      (*kcld3d)[iz][ix] = calloc((size_t) *Ny, sizeof(float)); 
      
      if (*gspec)
	(*gcld3d)[iz][ix] = calloc((size_t) *Ny, sizeof(float)); 

      if (*wspec)
	(*wcld3d)[iz][ix] = calloc((size_t) *Ny, sizeof(float)); 
    }
  }


  for (i=2; i<rows; i++) {

    number=0;
    while (number==0) {
      fgets (string, MAX_LENGTH_OF_LINE, f);
      status = ASCII_parsestring (string, &array, &number);
    
      if (status!=0) {
	fprintf (stderr, "Error %d reading 1st line of %s\n", status, filename);
	return status;
      }
    }

    if (!(*gspec))
      if (number<4) {
	fprintf (stderr, "Error: found %d columns in row %d, expected at least 4\n",
		 number, i);
	return -1;
      }

    if ((*gspec) && !(*wspec))
      if (number<5) {
	fprintf (stderr, "Error: found %d columns in row %d, expected at least 5\n",
		 number, i);
	return -1;
      }

    if (*gspec && *wspec)
      if (number<6) {
	fprintf (stderr, "Error: found %d columns in row %d, expected at least 6\n",
		 number, i);
	return -1;
      }
      
      
      
    /* cell indices */
    ix = strtol(array[0], &dummy, 0) - 1;
    iy = strtol(array[1], &dummy, 0) - 1;
    iz = strtol(array[2], &dummy, 0) - 1;

    if (ix<0 || ix>=*Nx) {
      fprintf (stderr, "Error, ix = %d out of bounds in line %d\n", ix+1, i+1);
      return -1;
    }

    if (iy<0 || iy>=*Ny) {
      fprintf (stderr, "Error, iy = %d out of bounds in line %d\n", iy+1, i+1);
      return -1;
    }

    if (iz<0 || iz>=*Nz) {
      fprintf (stderr, "Error, iz = %d out of bounds in line %d\n", iz+1, i+1);
      return -1;
    }


    /* copy data to result arrays */
    (*kcld3d)[iz][ix][iy] = strtod (array[3], &dummy);

    /* convert extinction from 1/km to 1/m */ 
    if (*gspec<3)
      (*kcld3d)[iz][ix][iy]/=1000.0;

    if (*gspec)
      (*gcld3d)[iz][ix][iy] = strtod (array[4], &dummy);

    if (*wspec)
      (*wcld3d)[iz][ix][iy] = strtod (array[5], &dummy);

    /* set flag indicating that this layer is 3D */
    (*threed)[iz] = 1;


    free(array);
  }


  /* close input file */
  fclose (f);

  if (!quiet)
    fprintf (stderr, " ... read %d data points from %s\n", 
	     rows-2, filename);

  return 0;
}





/*******************************************************/
/* Read data from a 2D albedo description file.        */
/*******************************************************/

int read_2D_albedo (char *filename, 
		    int *Nx, int *Ny,
		    double *delX, double *delY,
		    double ***albedo,
		    int quiet)
{
  int i=0, status=0;
  int ix=0, iy=0;

  double **value=NULL;
  int rows=0, min_columns=0, max_columns=0;


  /* read file; ASCII_file2double is a waste of memory in this */
  /* case, but it is very convenient because comments and      */
  /* everything are handled correctly.                         */

  status = ASCII_file2double (filename,
			      &rows, &max_columns, &min_columns, 
			      &value);
  
  if (status!=0) {
    fprintf (stderr, "Error %d reading file %s\n", status, filename);
    return status;
  }

  if (max_columns<4) {
    fprintf (stderr, "Error, found less than four columns in %s\n", 
	     filename);
    return -1;
  }
  
  if (rows<1) {
    if (!quiet)
      fprintf (stderr, "Error, no header in %s\n", filename);
    return -1;
  }


  /* 1st line, number of cells */
  *Nx = (int) (value[0][0] + 0.5);
  *Ny = (int) (value[0][1] + 0.5);


  /* 1st line, horizontal distances, convert from km to m */
  *delX = value[0][2] * 1000.0;
  *delY = value[0][3] * 1000.0;

  if (!quiet) {
    fprintf (stderr, " ... x-distance: %g\n", *delX);
    fprintf (stderr, " ... y-distance: %g\n", *delY);
  }

  /* allocate memory for albedo */
  *albedo = calloc((size_t) *Nx, sizeof(double *));

  for (ix=0; ix<*Nx; ix++)
    (*albedo)[ix] = calloc((size_t) *Ny, sizeof(double));


  for (i=1; i<rows; i++) {
    ix = (int) (value[i][0] + 0.5) - 1;
    iy = (int) (value[i][1] + 0.5) - 1;

    if (ix<0 || ix>=*Nx) {
      fprintf (stderr, "Error, ix = %d out of bounds in line %d\n", ix+1, i+1);
      return -1;
    }

    if (iy<0 || iy>=*Ny) {
      fprintf (stderr, "Error, iy = %d out of bounds in line %d\n", iy+1, i+1);
      return -1;
    }

    /* copy data to result arrays */
    (*albedo)[ix][iy] = value[i][2];

  }

  /* free memory */
  (void) ASCII_free_double (value, rows);
  

  if (!quiet)
    fprintf (stderr, " ... read %d data points from %s\n", 
	     rows-1, filename);

  return 0;
}


/*******************************************************/
/* Read data from a 2D RPV description file.           */
/*******************************************************/

int read_2D_rpv (char *filename, 
		 char **rpv_labels, int rpv_nlabels,
		 int *Nx, int *Ny,
		 double *delX, double *delY,
		 char ***rpv, int **rpv_count,
		 int quiet)
{
  int i=0, il=0, status=0;
  int ix=0, iy=0;

  int rows=0, min_columns=0, max_columns=0, max_length=0;

  char ***string=NULL;
  char *dummy=NULL;

#if HAVE_LIBNETCDF
  int start[1] = {0};
 
  int tstart[2] = {0,0};
  int tcount[2] = {0,0};
 
  int ncid=0, idd_nx=0, idd_ny=0;
  int id_delX=0, id_delY=0, id_type=0;

  unsigned char *temp=NULL;
  char tempstr[4]="";
#endif


  /* try open as netCDF file */
  
#if HAVE_LIBNETCDF
  status = nc_open (filename, NC_NOWRITE, &ncid);
  if (status==NC_NOERR) {

    if (!quiet)
      fprintf (stderr, " ... reading RPV data from netCDF file %s\n", filename);

    /* get dimension id for "nx" */
    status = nc_inq_dimid (ncid, "nx", &idd_nx);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading Nx from %s\n", status, filename);
      return status;
    }
    
    /* get dimension length for "nx" */
    status = nc_inq_dimlen (ncid, idd_nx, Nx);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading Nx from %s\n", status, filename);
      return status;
    }


    /* get dimension id for "ny" */
    status = nc_inq_dimid (ncid, "ny", &idd_ny);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading Ny from %s\n", status, filename);
      return status;
    }
    
    /* get dimension length for "ny" */
    status = nc_inq_dimlen (ncid, idd_ny, Ny);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading Ny from %s\n", status, filename);
      return status;
    }



    /* get variable id for "delX" */
    status = nc_inq_varid (ncid, "dx", &id_delX);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading delX from %s\n", 
	       status, filename);
      return status;
    }

    /* read "delX" */
    status = nc_get_var1_double (ncid, id_delX, start, delX);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading delX from %s\n", 
	       status, filename);
      return status;
    }


    /* get variable id for "delY" */
    status = nc_inq_varid (ncid, "dy", &id_delY);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading dy from %s\n", 
	       status, filename);
      return status;
    }

    /* read "delY" */
    status = nc_get_var1_double (ncid, id_delY, start, delY);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading dy from %s\n", 
	       status, filename);
      return status;
    }


    *delX *= 1000.0;
    *delY *= 1000.0;

    /* allocate memory for 2D BRDF field */
    *rpv = calloc((size_t) *Nx, sizeof(char *));
    *rpv_count = calloc((size_t) rpv_nlabels, sizeof(int));

    for (ix=0; ix<*Nx; ix++)
      (*rpv)[ix] = calloc((size_t) *Ny, sizeof(char));

    /* read data */

    /* get variable id for "type" */
    status = nc_inq_varid (ncid, "type", &id_type);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading type from %s\n", status, filename);
      return status;
    }

    temp = calloc (*Nx, sizeof(unsigned char));

    /* read type */
    tstart[1] = 0;   /* start with first x element    */
    tcount[1] = *Nx; /* read *Nx elements for each iy */

    tcount[0] = 1;  /* read one y element at a time */

    for (iy=0; iy<*Ny; iy++) {
      tstart[0] = iy;

      if (!quiet)
	fprintf (stderr, ".");
      
      status = nc_get_vara_uchar (ncid, id_type, tstart, tcount, temp);
      if (status!=NC_NOERR) {
	fprintf (stderr, "Error %d reading type from %s\n", status, filename);
	return status;
      }

      for (ix=0; ix<*Nx; ix++) {

	sprintf (tempstr, "%d", temp[ix]);
	
	for (il=0; il<rpv_nlabels; il++) {
	  if (!strcmp(tempstr, rpv_labels[il]))
	    break;
	}

	if (il==rpv_nlabels) {
	  fprintf (stderr, "Error, did not find an RPV entry for %s\n", tempstr);
	  return -1;
	}
	
	if (il>UCHAR_MAX) {
	  fprintf (stderr, "Error, RPV index %d larger than %d\n", 
		   il, UCHAR_MAX);
	  return -1;
	}
	  
	(*rpv)[ix][iy] = il; 
	(*rpv_count)[il]++;
      }
    }
    nc_close (ncid);

    free(temp);
    
    if (!quiet)
      fprintf (stderr, "\n");
  }
  else {
    if (!quiet)
      fprintf (stderr, " ... %s not in netCDF format, trying to open as ASCII\n", 
	       filename);
#endif
    
    /* read file; ASCII_file2double is a waste of memory in this */
    /* case, but it is very convenient because comments and      */
    /* everything are handled correctly.                         */
    
    status = ASCII_checkfile (filename, 
			      &rows,
			      &min_columns,
			      &max_columns,
			      &max_length);
    
    if (status!=0) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
    }
    
    if (max_columns<4) {
      fprintf (stderr, "Error, found less than four columns in %s\n", 
	       filename);
      return -1;
    }
    
    if (rows<1) {
      if (!quiet)
      fprintf (stderr, "Error, no header in %s\n", filename);
      return -1;
    }
    
    
    /* allocate memory */
    status = ASCII_calloc_string (&string,
				  rows,
				  max_columns,
				  max_length);
    
    if (status!=0) {
      fprintf (stderr, "Error %d allocating memory\n", status);
      return status;
    }
    
    
    /* read file to string array */
    status = ASCII_readfile (filename, string);
    
    if (status!=0) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
    }
    
  
    /* 1st line, number of cells */
    *Nx = strtol(string[0][0], &dummy, 0);
    *Ny = strtol(string[0][1], &dummy, 0);
    
    
    /* 1st line, horizontal distances, convert from km to m */
    *delX = strtod(string[0][2], &dummy) * 1000.0;
    *delY = strtod(string[0][3], &dummy) * 1000.0;
    
    if (!quiet) {
      fprintf (stderr, " ... x-distance: %g\n", *delX);
      fprintf (stderr, " ... y-distance: %g\n", *delY);
    }
    
    /* allocate memory for albedo */
    *rpv = calloc((size_t) *Nx, sizeof(char *));
    *rpv_count = calloc((size_t) rpv_nlabels, sizeof(int));

    for (ix=0; ix<*Nx; ix++)
      (*rpv)[ix] = calloc((size_t) *Ny, sizeof(char));
    
    
    for (i=1; i<rows; i++) {
      ix = strtol(string[i][0], &dummy, 0) - 1;
      iy = strtol(string[i][1], &dummy, 0) - 1;
      
      if (ix<0 || ix>=*Nx) {
	fprintf (stderr, "Error, ix = %d out of bounds in line %d\n", ix+1, i+1);
	return -1;
      }
      
      if (iy<0 || iy>=*Ny) {
	fprintf (stderr, "Error, iy = %d out of bounds in line %d\n", iy+1, i+1);
	return -1;
      }
      
      /* copy data to result arrays */
      for (il=0; il<rpv_nlabels; il++) {
	if (!strcmp(string[i][2], rpv_labels[il]))
	  break;
      }
      
      if (il==rpv_nlabels) {
	fprintf (stderr, "Error, did not find an RPV entry for %s\n", string[i][2]);
	return -1;
      }
      
      if (il>UCHAR_MAX) {
	fprintf (stderr, "Error, RPV index %d larger than %d\n", 
		 il, UCHAR_MAX);
	return -1;
      }
	  
      (*rpv)[ix][iy] = il; 
      (*rpv_count)[il]++;
    }

    /* free memory */
    (void) ASCII_free_string (string, rows, max_columns);

    if (!quiet)
      fprintf (stderr, " ... read %d data points from %s\n", 
	       rows-1, filename);

#if HAVE_LIBNETCDF
  }    
#endif

  return 0;
}




/*******************************************************/
/* Read data from a 2D elevation file.                 */
/*******************************************************/

int read_2D_elevation (char *filename, 
		       int *Nx, int *Ny,
		       double *delX, double *delY,
		       double ***elevation, int quiet)
{
  int i=0, status=0;
  int ix=0, iy=0;

  double **value=NULL;
  int rows=0, min_columns=0, max_columns=0;


  /* read file; ASCII_file2double is a waste of memory in this */
  /* case, but it is very convenient because comments and      */
  /* everything are handled correctly.                         */

  status = ASCII_file2double (filename,
			      &rows, &max_columns, &min_columns, 
			      &value);
  
  if (status!=0) {
    fprintf (stderr, "Error %d reading file %s\n", status, filename);
    return status;
  }

  if (max_columns<4) {
    fprintf (stderr, "Error: found less than four columns in %s\n", 
	     filename);
    return -1;
  }
  
  if (rows<1) {
    if (!quiet)
      fprintf (stderr, " ... found only one row in %s\n", filename);
    return -1;
  }


  /* 1st line, number of grid points */
  *Nx = (int) (value[0][0] + 0.5);
  *Ny = (int) (value[0][1] + 0.5);


  /* 1st line, horizontal distances, convert from km to m */
  *delX = value[0][2] * 1000.0;
  *delY = value[0][3] * 1000.0;

  if (!quiet) {
    fprintf (stderr, " ... x-distance: %g\n", *delX);
    fprintf (stderr, " ... y-distance: %g\n", *delY);
  }

  /* allocate memory for albedo */
  *elevation = calloc((size_t) *Nx, sizeof(double *));

  for (ix=0; ix<*Nx; ix++)
    (*elevation)[ix] = calloc((size_t) *Ny, sizeof(double));


  for (i=1; i<rows; i++) {
    ix = (int) (value[i][0] + 0.5) - 1;
    iy = (int) (value[i][1] + 0.5) - 1;

    if (ix<0 || ix>=*Nx) {
      fprintf (stderr, "Error, ix = %d out of bounds in line %d\n", ix+1, i+1);
      return -1;
    }

    if (iy<0 || iy>=*Ny) {
      fprintf (stderr, "Error, iy = %d out of bounds in line %d\n", iy+1, i+1);
      return -1;
    }

    /* copy data to result arrays, convert from km to m */
    (*elevation)[ix][iy] = value[i][2] * 1000.0;

  }

  /* free memory */
  (void) ASCII_free_double (value, rows);
  

  if (!quiet)
    fprintf (stderr, " ... read %d data points from %s\n", 
	     rows-1, filename);

  return 0;
}



/*******************************************************/
/* Read data from a 2D sampling grid file.             */
/*******************************************************/

int read_2D_sample (char *filename, 
		    int *Nx, int *Ny, int *Nz,
		    double *delX, double *delY, double **z,
		    double **theta, double **phi, double **alpha, int *nrad,
		    int quiet)
{
  int i=0, status=0;

  double **value=NULL;
  int rows=0, min_columns=0, max_columns=0;


  /* read file; ASCII_file2double is a waste of memory in this */
  /* case, but it is very convenient because comments and      */
  /* everything are handled correctly.                         */

  status = ASCII_file2double (filename,
			      &rows, &max_columns, &min_columns, 
			      &value);
  
  if (status!=0) {
    fprintf (stderr, "Error %d reading file %s\n", status, filename);
    return status;
  }

  if (max_columns<4) {
    fprintf (stderr, "Error: found less than four columns in %s\n", 
	     filename);
    return -1;
  }
  
  if (rows<2) {
    fprintf (stderr, "Error, found only one row in %s\n", filename);
    return -1;
  }


  /* 1st line, number of cells */
  *Nx = (int) (value[0][0] + 0.5);
  *Ny = (int) (value[0][1] + 0.5);


  /* 1st line, horizontal distances, convert from km to m */
  *delX = value[0][2] * 1000.0;
  *delY = value[0][3] * 1000.0;

  if (!quiet) {
    fprintf (stderr, " ... x-distance: %g\n", *delX);
    fprintf (stderr, " ... y-distance: %g\n", *delY);
  }

  /* 2nd line, number of altitudes */
  *Nz = (int) (value[1][0] + 0.5);

  if (max_columns < *Nz+1) {
    fprintf (stderr, "Error, found only %d columns in %s, expected %d\n", 
	     max_columns, filename, *Nz+1);
    return -1;
  }
  
  /* allocate memory for altitudes */
  if (*Nz>0)
    *z = calloc (*Nz, sizeof(double));
  
  /* 2nd line, altitudes [km], convert to m */
  for (i=0; i<*Nz; i++)
    (*z)[i] = value[1][1+i] * 1000.0;

  if (!quiet) {
    fprintf (stderr, " ... calculating result at the following %d altitudes:\n", *Nz);
    fprintf (stderr, " ... ");
  }

  if (!quiet) {
    for (i=0; i<*Nz; i++)
      fprintf (stderr, " %d", (int) ((*z)[i]+0.5));
    fprintf (stderr, "\n");
  }

  /* the following lines are angles for radiances */

  /* allocate memory for radiances */
  *nrad  = rows-2;

  if (*nrad>0) {
    *theta = calloc (*nrad, sizeof(double));
    *phi   = calloc (*nrad, sizeof(double));
    *alpha = calloc (*nrad, sizeof(double));
    
    for (i=0; i<*nrad; i++) {
      (*theta) [i] = value[i+2][0];
      (*phi)   [i] = value[i+2][1];
      (*alpha) [i] = value[i+2][2];
    }
  }

  /* free memory */
  ASCII_free_double (value, rows);

  return 0;
}




/***********************************************************************************/
/* Function: init_direction                                                        */
/* Description:                                                                    */
/*  Initialize direction with solar zenith sza and azimuth phi.                    */
/*  Convention:    phi=0,   South                                                  */
/*                 phi=90,  West                                                   */
/*                 phi=180, North                                                  */
/*                 phi=270, East                                                   */
/*  These definition refers to the position of the sun; e.g. phi=0 means that      */
/*  the sun is in the South and the photons are headed North.                      */
/*                                                                                 */
/* Parameters:                                                                     */
/* Return value:                                                                   */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                                 */
/***********************************************************************************/

void init_direction (double sinsza, double cossza, double sinphi, double cosphi, 
		     direction *dir)
{
  dir->dx[0]=  sinsza*sinphi;
  dir->dx[1]=  sinsza*cosphi;;
  dir->dx[2]= -cossza;

  /* cosine of solar zenith angle for radiance calculation */
  dir->cotheta = fabs(cossza);

  hitflag (dir);
}



/***********************************************************************************/
/* Function: hitflag                                                               */
/* Description:                                                                    */
/*  Calculate 'hitflag' which is used to determine the general direction of        */
/*  a photon:                                                                      */
/*   hit[i] =  1   positive                                                        */
/*   hit[i] =  0   negative                                                        */
/*   hit[i] = -1   no component                                                    */
/*                                                                                 */
/* Parameters:                                                                     */
/* Return value:                                                                   */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                                 */
/***********************************************************************************/

void hitflag (direction *dir)
{
  int i=0;

  for (i=0; i<3; i++) {
    dir->hit[i] = -1;
    
    if (dir->dx[i]>0)
      dir->hit[i]=1;
    else {
      if (dir->dx[i]<0)
	dir->hit[i] = 0;
    }
  }
}



