/*--------------------------------------------------------------------
 * $Id: albedo.c,v 1.5 2002/12/09 23:29:19 pa2h Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

#include <stdio.h>							    
#include <stdlib.h>
#include <string.h>
#include "uvspec.h"
#include "../src_c/ascii.h"



/************************************/
/* prototypes of internal functions */
/************************************/

static int read_albedo(char* albedo_filename, atm_out_struct *atm_out);
static int read_rpv (char *filename, rpv_out_struct *rpv);
static int interpolate_rpv (rpv_out_struct *rpv, float *lambda, int nlambda);
static int read_surfaces (char *filename, surfaces_out_struct *surf,
			  float *lambda, int nlambda);


/**************************************************************/
/* Setup surface albedo.                                      */
/**************************************************************/

int setup_albedo (input_struct input, output_struct *output)
{
  int status=0, iv=0, linear=0;

  output->atm.albedo_t = (float *) calloc (output->wl.nlambda_t, sizeof(float));

  output->rpv.rho0_t   = (float *) calloc (output->wl.nlambda_t, sizeof(float));
  output->rpv.k_t      = (float *) calloc (output->wl.nlambda_t, sizeof(float));
  output->rpv.theta_t  = (float *) calloc (output->wl.nlambda_t, sizeof(float));


  if (input.atm.albedo < 0.0) {
    status = read_albedo (input.filename[FN_ALBEDO], &output->atm);
    if (status != 0) {
      fprintf (stderr, "Error reading albedo file: %s\n", input.filename[FN_ALBEDO]);
      return status;
    }
    
    /**** Interpolate albedo to transmission grid ****/
    
    linear = 1;  /**** Use linear interpolation for the albedo, otherwise 
		       negative albedos or albedos larger than 1 may occur, 
		       I should know:-) AKy, ****/ 

    status = arb_wvn (output->atm.file_nlambda_albedo, output->atm.file_lambda_albedo,
		      output->atm.file_albedo, 
		      output->wl.nlambda_t, output->wl.lambda_t, output->atm.albedo_t, 
		      linear, 0);

    if (status != 0) {
      fprintf (stderr, "Error %d returned by arb_wvn()\n", status);
      return status;
    }
    
    /**** Check interpolated albedo values ****/

    for (iv=0; iv<output->wl.nlambda_t; iv++) {
      if (output->atm.albedo_t[iv] > 1.0 || output->atm.albedo_t[iv] < 0.0)  {
	fprintf (stderr, "Albedo interpolated from %s out of range:\n", input.filename[FN_ALBEDO]);
	fprintf (stderr, "wavelength: %f, albedo %f\n", 
		 output->wl.lambda_t[iv], output->atm.albedo_t[iv]);
	fprintf (stderr, "Modify albedo file %s\n", input.filename[FN_ALBEDO]);
	return -1;
      } 
    }
  }
  else
    for (iv=0;iv<output->wl.nlambda_t;iv++)
      output->atm.albedo_t[iv] = input.atm.albedo;    
  

  /* read BRDF */
  if (strlen(input.filename[FN_RPV]) > 0) {
    
    status = read_rpv (input.filename[FN_RPV], &(output->rpv));
    if (status!=0) {
      fprintf (stderr, "Error %d returned by read_rpv (%s)\n", status, input.filename[FN_RPV]);
      return status;
    }
    

    status = interpolate_rpv (&(output->rpv), output->wl.lambda_t, output->wl.nlambda_t);
    if (status!=0) {
      fprintf (stderr, "Error %d returned by interpolate_rpv()\n", status);
      return status;
    }
  }
  else {
    for (iv=0;iv<output->wl.nlambda_t;iv++) {
      output->rpv.rho0_t  [iv] = input.rpv.rho0;    
      output->rpv.k_t     [iv] = input.rpv.k;    
      output->rpv.theta_t [iv] = input.rpv.theta;    
    }
  }


  /* read surface information for MYSTIC */
  if (strlen(input.rte.mc.filename[FN_MC_RPV]) > 0) {

    status = read_surfaces(input.rte.mc.filename[FN_MC_RPVTYPE], &(output->surf),
			   output->wl.lambda_t, output->wl.nlambda_t);
    if (status!=0) {
      fprintf (stderr, "Error %d returned by read_surfaces(%s)\n", status, input.rte.mc.filename[FN_MC_RPV]);
      return status;
    }

  }    

  return 0;
}





/**************************************************************/
/* Read 2-column albedo file.                                 */
/**************************************************************/

static int read_albedo (char *filename, atm_out_struct *atm_out) 
{
  int status=0;
  
  status = read_2c_file_float (filename, 
			       &(atm_out->file_lambda_albedo),
			       &(atm_out->file_albedo),
			       &(atm_out->file_nlambda_albedo));

  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }

  return 0;
}


/**************************************************************/
/* Read 4-column RPV file.                                    */
/**************************************************************/

static int read_rpv (char *filename, rpv_out_struct *rpv)
{
  int status=0;

  status = read_4c_file_float (filename,
			       &(rpv->lambda),
			       &(rpv->rho0),
			       &(rpv->k),
			       &(rpv->theta),
			       &(rpv->n));
  
  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }

  return 0; /* if o.k. */
}  



/**************************************************************/
/* Interpolate RPV to internal wavelength grid.               */
/**************************************************************/

static int interpolate_rpv (rpv_out_struct *rpv, float *lambda, int nlambda)
{
  int status=0;

  int linear=1;   /* linear interpolation */

  /* interpolate to internal RTE grid */
  status = arb_wvn (rpv->n, rpv->lambda, rpv->rho0, 
		    nlambda, lambda, rpv->rho0_t, 
		    linear, 0);
  
  if (status != 0) {
    fprintf (stderr, "Error %d returned by arb_wvn(rho0)\n", status);
    return status;
  }
  
  status = arb_wvn (rpv->n, rpv->lambda, rpv->k, 
		    nlambda, lambda, rpv->k_t, 
		    linear, 0);
  
  if (status != 0) {
    fprintf (stderr, "Error %d returned by arb_wvn(k)\n", status);
    return status;
  }
  
  status = arb_wvn (rpv->n, rpv->lambda, rpv->theta, 
		    nlambda, lambda, rpv->theta_t, 
		    linear, 0);
  
  if (status != 0) {
    fprintf (stderr, "Error %d returned by arb_wvn(theta)\n", status);
    return status;
  }
  
  return 0;  /* if o.k. */
}
  
  

/**************************************************************/
/* Read surfaces from file.                                   */
/**************************************************************/

static int read_surfaces (char *filename, surfaces_out_struct *surf, float *lambda, int nlambda) 
{
  int il=0, status=0;
  int rows=0, min_columns=0, max_columns=0, max_length=0;
  char ***string=NULL;

  status = ASCII_checkfile (filename, 
			    &rows,
			    &min_columns,
			    &max_columns,
			    &max_length);
  
  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }

  if (min_columns<2)
    fprintf (stderr, "Error, found less than 2 columns in %s\n", filename);
  
  if (min_columns != max_columns)
    fprintf (stderr, "Warning, inconsistent number of columns in %s\n", filename);

  status = ASCII_calloc_string (&string,
				rows,
				max_columns,
				max_length);

  if (status!=0) {
    fprintf (stderr, "Error %d allocating memory\n", status);
    return status;
  }

  status = ASCII_readfile (filename, string);

  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }

  surf->n = rows;
  surf->label    = calloc (surf->n, sizeof (char *));
  surf->filename = calloc (surf->n, sizeof (char *));
  surf->rpv      = calloc (surf->n, sizeof (rpv_out_struct));

  for (il=0; il<surf->n; il++) {
    surf->label[il]    = calloc (max_length+1, sizeof(char));
    surf->filename[il] = calloc (max_length+1, sizeof(char));

    strcpy (surf->label[il], string[il][0]);
    strcpy (surf->filename[il], string[il][1]);
  }
  
  ASCII_free_string (string, rows, max_columns);

  /* now read filenames and interpolate data to internal resolution */
  for (il=0; il<surf->n; il++) {
    fprintf (stderr, " ... reading type %s from %s\n", surf->label[il], surf->filename[il]);

    /* read wavelength-dependent RPV data */
    status = read_rpv (surf->filename[il], &(surf->rpv[il]));
    if (status!=0) {
      fprintf (stderr, "Error %d returned by read_rpv(%s)\n", status, surf->filename[il]);
      return status;
    }

    /* allocate memory */
    surf->rpv[il].rho0_t   = (float *) calloc (nlambda, sizeof(float));
    surf->rpv[il].k_t      = (float *) calloc (nlambda, sizeof(float));
    surf->rpv[il].theta_t  = (float *) calloc (nlambda, sizeof(float));

    /* interpolate to internal wavelength grid */
    status = interpolate_rpv (&(surf->rpv[il]), lambda, nlambda);
    if (status!=0) {
      fprintf (stderr, "Error %d returned by interpolate_rpv()\n", status);
      return status;
    }
  }

  return 0;
}
